// @flow
'use strict'

import type { Message } from '../types'
import type { request } from 'restify/lib/request'
import type { response } from 'restify/lib/response'

const _ = require('lodash')
const joi = require('joi')
const logger = require('../lib/logger').queueLogger
const errors = require('restify-errors')
const sqsConsumer = require('../lib/sqs-consumer')
const schema = require('./result-processor-schema')
const {
  mergeAnalysisResults,
  resendTaskOutput
} = require('../result-processor/result-processor-service')
const config = require('config')
const errorHandler = require('../lib/error-handler')
const taskService = require('../task/task-service')
const { TASK_STAGE } = require('../constant')

const batchEventResponseQueue = config.get('batchEventResponseQueue')

const JOB_NAME = {
  series_extraction_qc: 'cleerly-job-moz-series-extraction',
  series_ranking: 'cleerly-job-moz-series-ranking',
  central_line_detection: 'cleerly-job-moz-artery-tree',
  cmpr: 'cleerly-job-moz-cmpr-generation',
  attraction_points: 'cleerly-job-moz-attraction-point',
  variable_analysis: 'cleerly-job-moz-variable-generation'
}

async function handler (message: Message) {
  logger.info(`Received a result processor message: ${message.Body}`)

  const messageBody = JSON.parse(message.Body)
  joi.assert(messageBody, schema)
  const { taskId, studyId, series, status } = messageBody.subtaskResult
  return mergeAnalysisResults(taskId, studyId, series, status, messageBody.subtaskOutputData)
}

module.exports = {
  batchEventResponseConsumer: sqsConsumer(batchEventResponseQueue, handler),

  handler,

  async rerunResultProcessor (req: request, res: response, next: Function) {
    const studyId = req.query.studyId
    const taskId = req.query.taskId
    const stage = req.query.stage

    try {
      let task
      if (studyId && !taskId) {
        task = await taskService.getLastSeriesExtractionTaskForStudy(studyId)
      }
      if (taskId) {
        task = await taskService.getTaskById(taskId)
      }

      if (!task || !task.success) {
        return next(new errors.NotFoundError(`Study ${studyId} task ${taskId} is not found or the task is failed`))
      }

      let subtaskStages = []
      if (stage && JOB_NAME[stage]) {
        subtaskStages.push(stage)
      } else {
        switch (task.stage) {
          case TASK_STAGE.SERIES_EXTRACTION_QC:
            subtaskStages.push(TASK_STAGE.SERIES_EXTRACTION_QC)
          case TASK_STAGE.SERIES_RANKING: // eslint-disable-line no-fallthrough
            subtaskStages.push(TASK_STAGE.SERIES_RANKING)
          case TASK_STAGE.CENTRAL_LINE_DETECTION: // eslint-disable-line no-fallthrough
            subtaskStages.push(TASK_STAGE.CENTRAL_LINE_DETECTION)
          case TASK_STAGE.CMPR: // eslint-disable-line no-fallthrough
            subtaskStages.push(TASK_STAGE.CMPR)
          case TASK_STAGE.ATTRACTION_POINTS: // eslint-disable-line no-fallthrough
            subtaskStages.push(TASK_STAGE.ATTRACTION_POINTS)
          case TASK_STAGE.VARIABLE_ANALYSIS: // eslint-disable-line no-fallthrough
            subtaskStages.push(TASK_STAGE.VARIABLE_ANALYSIS)
        }
      }

      if (_.isEmpty(subtaskStages)) {
        return next(new errors.BadRequestError(`Task ${task.id} has no stage`))
      }

      await resendTaskOutput(task, subtaskStages)
      await taskService.updateTask(task.id, {
        success: false,
        date_completed: null
      })

      res.json({
        message: `Rerun result processor for study (${task.studyId}) - task (${task.id}): ${subtaskStages.join(', ')}`
      })
      return next()
    } catch (err) {
      return errorHandler.onError(req, res, next)(err)
    }
  }
}
