// @flow
'use strict'

import type { Series, Task, Stage, PersistedVessel } from '../types'

const _ = require('lodash')
const shortId = require('shortid')
const config = require('config')
const logger = require('../lib/logger').queueLogger
const {
  setBestCombination,
  mergeSeriesList,
  clearVesselPendingFlag,
  clearSeriesPendingFlag,
  deriveStatus
} = require('./result-processor-util')
const studyService = require('../study/study-service')
const taskService = require('../task/task-service')
const db = require('../lib/db')
const AWS = require('../lib/aws')
const { TASK_STAGE, STUDY_STATUS } = require('../constant')
const { tableColumn: reviewColumn } = require('@cleerly/coronary-db-schema/study-review')

const batchEventQueue = config.get('batchEventQueue')
const batchEventResponseQueue = config.get('batchEventResponseQueue')
const mozartResultBucketName = config.get('mozartResultBucket.bucketName')

const JOB_NAME = {
  series_extraction_qc: 'cleerly-job-moz-series-extraction',
  series_ranking: 'cleerly-job-moz-series-ranking',
  central_line_detection: 'cleerly-job-moz-artery-tree',
  cmpr: 'cleerly-job-moz-cmpr-generation',
  attraction_points: 'cleerly-job-moz-attraction-point',
  variable_analysis: 'cleerly-job-moz-variable-generation'
}

module.exports = {
  async mergeAnalysisResults (
    taskId: string,
    studyId: string,
    series: Array<Series>,
    status: number,
    subtaskOutputData: Object
  ) {
    await db.transaction(async trx => {
      const study = await studyService.getStudyByIdForUpdate(studyId, trx)
      let studyReview = await studyService.getMostRecentStudyReviewByStudyId(studyId, trx) || {}
      const task = await taskService.getTaskByIdForUpdate(taskId, trx)

      const taskStage = task.stage
      if (task.success) {
        logger.info(`Task ${taskId} ${taskStage} of study ${studyId} has already been completed. Ignore the message`)
        return
      }

      const mergedSeries = await mergeSeriesList(studyReview.series, series, studyId)
      const mergedErrors = _.compact(_.concat(studyReview.errors, subtaskOutputData.errors))
      const mergedOutputData = { ...task.outputData, ...subtaskOutputData }
      const subtaskStatus = deriveStatus(task.subtaskStatus, status)

      const studyUpdate = {}
      const studyReviewUpdate = {}
      if (studyReview.id) {
        studyReviewUpdate.id = studyReview.id
      } else {
        studyReviewUpdate[reviewColumn.reviewedByMozart] = true
      }

      // If the series extraction step failed, make the study as failed
      studyReviewUpdate.errors = JSON.stringify(mergedErrors)

      const taskUpdate = {}
      taskUpdate.output_data = JSON.stringify(mergedOutputData)
      taskUpdate.subtask_status = subtaskStatus

      if (_.some(subtaskOutputData.errors, ['task', TASK_STAGE.SERIES_EXTRACTION_QC])) {
        // If the series extraction step fail, the study analysis failed
        studyUpdate.status = STUDY_STATUS.ANALYSIS_FAILED
      }

      // Task is completed
      if (subtaskStatus === 0) {
        logger.info(`Task ${taskId} ${taskStage} of study ${studyId} is completed.`)

        if (study.status === STUDY_STATUS.PENDING ||
          study.status === STUDY_STATUS.NEW) {
          studyUpdate.status = STUDY_STATUS.NOT_IN_REVIEW
        }

        studyUpdate.date_analyzed = trx.fn.now()
        taskUpdate.success = true
        taskUpdate.date_completed = trx.fn.now()

        if (taskStage === TASK_STAGE.SERIES_EXTRACTION_QC) {
          // mark best combination vessels
          setBestCombination(_.orderBy(mergedSeries, 'ranking', 'desc'))
        } else if (taskStage === TASK_STAGE.CENTRAL_LINE_DETECTION) {
          clearSeriesPendingFlag(mergedSeries, series)
        } else if (taskStage === TASK_STAGE.CMPR) {
          // clear vessel not analyzed flag
          clearVesselPendingFlag(mergedSeries, series)
        }
      }
      studyReviewUpdate.series = JSON.stringify(mergedSeries)

      await studyService.updateStudyById(studyId, studyUpdate, trx)
      if (studyReviewUpdate.id) {
        studyService.updateStudyReviewById(studyReviewUpdate.id, studyReviewUpdate, trx)
      } else {
        await studyService.createOrUpdateStudyReview(taskId, studyId, studyReviewUpdate, trx)
      }
      await taskService.updateTask(taskId, taskUpdate, trx)
    })
  },

  async resendTaskOutput (task: Task, stages: Array<Stage>) {
    const sqs = new AWS.SQS({
      region: batchEventResponseQueue.region
    })

    await sqs.sendMessageBatch({
      Entries: stages.map(sta => {
        // CloudWatch event message body parsed in result processor
        return {
          Id: `${task.id}-${shortId.generate()}`,
          MessageBody: JSON.stringify({
            source: 'visionlab.api',
            time: new Date().toISOString(),
            detail: {
              status: 'SUCCEEDED',
              parameters: {
                taskType: 'SUBTASK',
                taskStage: task.stage,
                response: `s3://${mozartResultBucketName}/${task.id}/${sta}/response.json`,
                taskName: JOB_NAME[sta],
                studyId: task.studyId,
                taskId: task.id
              }
            }
          })
        }
      }),
      QueueUrl: batchEventQueue.url
    }).promise()
  },

  async sendStudyReview (reportId: string, vessels: Array<PersistedVessel>) {
    const sqs = new AWS.SQS({
      region: batchEventResponseQueue.region
    })

    await sqs.sendMessage({
      MessageBody: JSON.stringify({
        type: 'REPORT_GENERATION',
        reportId,
        vessels
      }),
      QueueUrl: batchEventQueue.url
    }).promise()
  }
}
