// @flow
'use strict'

import type { PersistedSeries, Series } from '../types'

const _ = require('lodash')
const shortId = require('shortid')
const { isUnknownLabel } = require('../vessel/vessel-label-util')
const { toObject } = require('../series/series-util')
const { getS3Json, putS3Json } = require('../lib/s3-helper')
const config = require('config')

const resultBucket = config.get('mozartResultBucket')

/**
 * Mark the best combination series vessels from the series extraction Mozart result
 *
 * This function mutates seriesList
 * @param seriesList
 */
function setBestCombination (seriesList: Array<PersistedSeries>) {
  const vesselLabelKeys = {}
  for (const series of seriesList) {
    _.forEach(series.vessels, vessel => {
      const labelKey = vessel.label
      if (!isUnknownLabel(labelKey) && !vesselLabelKeys[labelKey]) {
        vessel.isBestCombination = true
        vesselLabelKeys[labelKey] = true
      } else {
        delete vessel.isBestCombination
      }
    })
  }
}

/**
 * Merge customizer
 * If customizer returns undefined, merging is handled by the method instead
 * @param objValue
 * @param srcValue
 * @returns {*}
 */
function mergeCopyArray (objValue: mixed, srcValue: mixed) {
  if (_.isArray(objValue)) {
    return srcValue
  }
}

/**
 * Clear the series pending flag
 * This function mutate the seriesDict
 * @param seriesDict
 * @param analyzedSeries
 */
function clearSeriesPendingFlag (seriesDict: { [string]: PersistedSeries }, analyzedSeries: Array<Series>) {
  _.forEach(analyzedSeries, s => {
    if (seriesDict[s.id]) {
      delete seriesDict[s.id].notAnalyzed
      delete seriesDict[s.id].pending
    }
  })
}

/**
 * Clear the marks of the not analyzed vessels
 *
 * This function mutate the seriesDict input parameter
 * @param seriesDict
 * @param analyzedSeries
 */
function clearVesselPendingFlag (seriesDict: { [string]: PersistedSeries }, analyzedSeries: Array<Series>) {
  for (const series of analyzedSeries) {
    const targetSeries = seriesDict[series.id]
    _.forEach(series.vessels, vessel => {
      const targetVessel = targetSeries.vessels[vessel.id]
      delete targetVessel.notAnalyzed
      delete targetVessel.pending
    })
  }
}

/**
 * Merge the series list and it's artery tree vessels
 *
 * This function does not mutate the seriesList.
 * @param seriesObject
 * @param incomingSeriesList
 * @param studyId
 * @return merged series list
 */
async function mergeSeriesList (
  seriesObject: ?{ [key:string]: Series },
  incomingSeriesList: Array<Series>,
  studyId: string
): Promise<{ [string]: PersistedSeries }> {
  await Promise.all(
    incomingSeriesList.map(async (incomingSeries: Series) => {
      if (!seriesObject) {
        seriesObject = {}
      }
      const srcSeries = seriesObject[incomingSeries.id]

      // Set notAnalyzed flag if the incoming series has no vessels
      incomingSeries.notAnalyzed = _.isEmpty(incomingSeries.vessels)

      if (incomingSeries.centralLineRefFile && srcSeries && srcSeries.centralLineRefFile) {
        const [incomingArteryTree, originalArteryTree] = await Promise.all([
          getS3Json(incomingSeries.centralLineRefFile),
          getS3Json(srcSeries.centralLineRefFile)
        ])

        const incomingVessels = _.keyBy(incomingArteryTree.vessels, 'id')
        const originalVessels = _.keyBy(originalArteryTree.vessels, 'id')

        const mergedArteryTree = {
          vessels: _.values({ ...originalVessels, ...incomingVessels }, 'id')
        }

        const mergedCentralLineRefFile = `s3://${resultBucket.bucketName}/${studyId}/${shortId.generate()}/${srcSeries.id}/merged-artery-tree.json`
        await putS3Json(mergedCentralLineRefFile, mergedArteryTree)

        // Reset incoming series list central line ref file
        incomingSeries.centralLineRefFile = mergedCentralLineRefFile
      }
    })
  )

  return _.mergeWith(seriesObject, toObject(incomingSeriesList), mergeCopyArray)
}

/**
 * Check derive status to see if the task is completed.
 *
 * each bit of number status stands for the status of subtask.
 * 0 means processed, 1 means not processed. Max number of bits is 32 (4 bytes).
 *
 * for example,
 * status of subtask series extraction is 0x111110
 * status of subtask sereis ranking is 0x111101
 * status of subtask artery tree generation is 0x111011,
 * status of subtask cmpr generation is 0x110111
 * status of subtask attraction point generation is 0x101111
 * status of subtask variable generation is 0x011111
 *
 * when currentStatus equals 0, all subtasks of the task have been processed.
 * the purpose of using bit is to decouple the number of subtasks from the status of task.
 * the status of the task equals 0 when all subtasks are processed regardless the number of subtasks.
 *
 * @param currentStatus
 * @param status
 * @returns {number}
 */
function deriveStatus (currentStatus: number, status: number) {
  if (_.isNil(status)) {
    return currentStatus
  }

  let mergedStatus = status
  if (!_.isNil(currentStatus)) {
    mergedStatus = currentStatus & status
  }
  return mergedStatus
}

module.exports = {
  setBestCombination,
  mergeSeriesList,
  clearVesselPendingFlag,
  clearSeriesPendingFlag,
  deriveStatus
}
