// @flow
'use strict'

import type { request } from 'restify/lib/request'
import type { response } from 'restify/lib/response'

const _ = require('lodash')
const errors = require('restify-errors')
const db = require('../lib/db')
const url = require('config').get('url')
const utils = require('../utils')
const errorHandler = require('../lib/error-handler')
const studyService = require('../study/study-service')
const seriesService = require('./series-service')
const { createTaskForSeriesArteryTreeAnalysis } = require('../task/task-service')
const { getSeries } = require('./series-util')
const { sendTaskSQSMessage } = require('../task/task-service')

module.exports = {
  async getSeriesById (req: request, res: response, next: Function) {
    const studyId: string = req.params.studyId
    const seriesId: string = req.params.seriesId

    try {
      const series = await seriesService.getSeriesById(studyId, seriesId)

      if (!series) {
        return next(new errors.NotFoundError(`Series with id ${studyId} ${seriesId} is not found`))
      }

      res.json(getSeries(series))
      return next()
    } catch (err) {
      return errorHandler.onError(req, res, next)(err)
    }
  },

  async getSeriesViewer (req: request, res: response, next: Function) {
    const studyId: string = req.params.studyId
    const seriesId: string = req.params.seriesId

    try {
      const series = await seriesService.getSeriesById(studyId, seriesId)

      if (!series) {
        return next(new errors.NotFoundError(`Series with id ${studyId} ${seriesId} is not found`))
      }

      if (!series.frames) {
        return next(new errors.NotFoundError(`Frames is not found in series ${studyId} ${seriesId}`))
      }

      const axialPrefix = `${url.content}/${utils.getS3InfoFromURI(series.frames.axial.prefix).key}`
      const coronalPrefix = `${url.content}/${utils.getS3InfoFromURI(series.frames.coronal.prefix).key}`
      const sagittalPrefix = `${url.content}/${utils.getS3InfoFromURI(series.frames.sagittal.prefix).key}`

      const frameType = _.get(series.frames, 'metadata.frameType', 'png')
      const metadata = { ...series.metadata, ...series.frames.metadata }
      const frames = {
        metadata,
        axial: {
          width: series.frames.axial.width,
          height: series.frames.axial.height,
          frames: _.range(series.frames.axial.imageCount).map(n => `${axialPrefix}${n}.${frameType}`)
        },
        coronal: {
          width: series.frames.coronal.width,
          height: series.frames.coronal.height,
          frames: _.range(series.frames.coronal.imageCount).map(n => `${coronalPrefix}${n}.${frameType}`)
        },
        sagittal: {
          width: series.frames.sagittal.width,
          height: series.frames.sagittal.height,
          frames: _.range(series.frames.sagittal.imageCount).map(n => `${sagittalPrefix}${n}.${frameType}`)
        }
      }

      res.json(frames)
      return next()
    } catch (err) {
      return errorHandler.onError(req, res, next)(err)
    }
  },

  async analyzeSeries (req: request, res: response, next: Function) {
    const studyId: string = req.params.studyId
    const seriesId: string = req.params.seriesId
    const userId = req.user.sub

    try {
      let task
      let series = {}
      await db.transaction(async trx => {
        let studyReview = await studyService.getStudyReviewByStudyIdForUpdate(studyId, trx) || {}
        if (studyReview.reviewedByMozart) {
          studyReview = await studyService.createStudyReviewFromMozartResult(studyId, userId, trx)
        }

        const result = await seriesService.getSeriesByIdForUpdate(studyReview.id, seriesId, trx)
        series = result.series
        if (!series) {
          return next(new errors.NotFoundError(`Series with id ${studyId} ${seriesId} is not found`))
        }

        task = await createTaskForSeriesArteryTreeAnalysis(studyId, series, trx)

        // Add pending flag on the series
        series.pending = true
        series.notAnalyzed = true
        await seriesService.updateSeries(studyReview.id, seriesId, series, trx)
      })
      // Send task message after the db transaction
      if (task) {
        await sendTaskSQSMessage(task)
      }

      res.json(getSeries(series))
      return next()
    } catch (err) {
      return errorHandler.onError(req, res, next)(err)
    }
  }
}
