// @flow
'use strict'

import type { PersistedVessel, Statistics, PlaqueAreas } from '../types'

const _ = require('lodash')
const https = require('https')
const axios = require('axios')
const config = require('config')
const { requestTimeout } = require('../constant')
const { CENTRAL_LINE_SPACING } = require('./vessel-constant')
const { getCTOArea } = require('./vessel-util')

const resultProcessorServiceUrl: string = config.get('services.resultProcessorServiceUrl')

const axiosInstance = axios.create({
  baseURL: `${resultProcessorServiceUrl}/api/v1`,
  timeout: requestTimeout,
  httpsAgent: new https.Agent({ keepAlive: true })
})

module.exports = {
  axiosInstance,

  async getStatisticsPlaqueAreas (
    studyId: string,
    seriesId: string,
    vesselId: string,
    vessel: PersistedVessel,
    centralLineSpacing: number = CENTRAL_LINE_SPACING
  ): Promise<{
    statistics: Statistics,
    plaqueAreas?: PlaqueAreas,
    variablesRefFile?: string
  }> {
    let requestPayload = {}
    const payload = {
      studyId,
      seriesId,
      vesselId,
      label: vessel.label,
      start: vessel.start,
      end: vessel.end,
      areas: vessel.areas,
      centralLineSpacing
    }

    let lumenWall = _.cloneDeep(vessel.lumenWall)
    if (lumenWall) {
      if (vessel.areas && vessel.areas.ALL && vessel.lumenWall) {
        const ctoArea = getCTOArea(vessel.areas.ALL)
        // Set the lumen to be a small circle in the CTO area
        if (ctoArea && _.isArray(ctoArea.sections)) {
          for (const { start, end } of ctoArea.sections) {
            for (let i = start; i <= end; i++) {
              lumenWall[i] = {
                '0': [{ x: 31.4999, y: 31.5 }, { x: 31.5001, y: 31.5 }],
                '22.5': [{ x: 31.49999, y: 31.50001 }, { x: 31.50001, y: 31.49999 }],
                '45': [{ x: 31.4999, y: 31.5001 }, { x: 31.5001, y: 31.4999 }],
                '67.5': [{ x: 31.499, y: 31.501 }, { x: 31.501, y: 31.499 }],
                '90': [{ x: 31.5, y: 31.5001 }, { x: 31.5, y: 31.4999 }],
                '112.5': [{ x: 31.50001, y: 31.5001 }, { x: 31.4999, y: 31.49999 }],
                '135': [{ x: 31.5001, y: 31.5001 }, { x: 31.4999, y: 31.4999 }],
                '157.5': [{ x: 31.5001, y: 31.50001 }, { x: 31.49999, y: 31.4999 }]
              }
            }
          }
        }
      }
      requestPayload = {
        ...payload,
        cmprRefFile: vessel.cmprRefFile,
        lumenWall: lumenWall,
        vesselWall: vessel.vesselWall
      }
    } else if (vessel.variablesRefFile) {
      requestPayload = {
        ...payload,
        variablesRefFile: vessel.variablesRefFile
      }
    } else {
      throw new Error(`Study ${studyId} series ${seriesId} vessel ${vesselId} has no lumen vessel wall or variable ref file when calling statistics service.`)
    }

    const response = await axiosInstance.post('/statistics/', _.omitBy(requestPayload, _.isNil))
    return response.data
  }
}
