// @flow
'use strict'

import type { request } from 'restify/lib/request'
import type { response } from 'restify/lib/response'
import type { Wall, PersistedVessel } from '../types'

const _ = require('lodash')
const joi = require('joi')
const sentry = require('@sentry/node')
const errors = require('restify-errors')
const shortId = require('shortid')
const NodeCache = require('node-cache')
const streamToPromise = require('stream-to-promise')
const db = require('../lib/db')
const { dicomBucket } = require('config')
const { CENTRAL_LINE_SPACING } = require('../vessel/vessel-constant')
const s3Helper = require('../lib/s3-helper')
const errorHandler = require('../lib/error-handler')
const { getStudyByIdForUpdate, updateStudyReviewById } = require('../study/study-service')
const { addPendingOnVessels } = require('../study/study-util')
const studyService = require('../study/study-service')
const seriesService = require('../series/series-service')
const vesselService = require('./vessel-service')
const { createTaskForVesselsCMPRAnalysis } = require('../task/task-service')
const studyUtil = require('../study/study-util')
const vesselUtil = require('./vessel-util')
const { sendTaskSQSMessage } = require('../task/task-service')
const { getStudyPermission } = require('../study/study-util')
const { manualHelper, preloadCentralPoint, adjustCentralPoint } = require('./helper-service')
const { getStatisticsPlaqueAreas } = require('./statistics-plaque-service')

const manualHelperPathCache = new NodeCache({
  stdTTL: 600 // 10 minutes
})
const centralPointCache = new NodeCache({
  stdTTL: 6000 // 100 minutes
})

const manualHelperKey = (start, end, seriesPath) => {
  return `[${start.x},${start.y},${start.z}]/[${end.x},${end.y},${end.z}]/${seriesPath}`
}

/**
 * This function mutate the old full wall
 * @param oldFullWall
 * @param startIndex
 * @param wallSection
 * @returns {Array}
 */
function getFullWall (oldFullWall: Array<Wall>, startIndex: number, wallSection: ?Array<Wall>): Array<Wall> {
  if (!wallSection || _.isEmpty(wallSection)) {
    return oldFullWall
  }

  for (let i = startIndex, len = oldFullWall.length; i < len && i - startIndex < wallSection.length; i++) {
    oldFullWall[i] = {
      ...oldFullWall[i],
      ...wallSection[i - startIndex]
    }
  }
  return oldFullWall
}

async function getWalls (vessel: PersistedVessel, wall: {
  startIndex: number,
  lumen: Array<Wall>,
  vessel: Array<Wall>
}) {
  if (wall && _.isNil(wall.startIndex)) {
    return {
      lumenWall: wall.lumen,
      vesselWall: wall.vessel
    }
  }

  const wallPromises = []
  wallPromises.push(s3Helper.getS3Json(vessel.lumenWallRefFile))
  wallPromises.push(s3Helper.getS3Json(vessel.vesselWallRefFile))

  const [lumenWall, vesselWall] = await Promise.all(wallPromises)
  if (wall && !_.isNil(wall.startIndex)) {
    getFullWall(lumenWall, wall.startIndex, wall.lumen)
    getFullWall(vesselWall, wall.startIndex, wall.vessel)
  }

  return {
    lumenWall,
    vesselWall
  }
}

module.exports = {
  async getArteryTree (req: request, res: response, next: Function) {
    const studyId: string = req.params.studyId
    const seriesId: string = req.params.seriesId

    try {
      const series = await seriesService.getSeriesById(studyId, seriesId)

      if (!series) {
        return next(new errors.NotFoundError(`Series with id ${studyId} ${seriesId} is not found`))
      }

      const arteryTree = await s3Helper.getS3Json(series.centralLineRefFile)

      let vessels = []
      if (!_.isEmpty(arteryTree)) {
        const vesselDict = _.keyBy(arteryTree.vessels, 'id')
        vessels = _(series.vessels)
          .map(vessel => {
            vessel.vertexes = vesselDict[vessel.id] && vesselDict[vessel.id].vertexes
            return vesselUtil.getVesselSummary(vessel)
          })
          .sort(vesselUtil.vesselComparator)
          .value()
      }

      const result = {}
      result.vessels = vessels

      if (series.aortaMesh) {
        result.aortaMesh = vesselUtil.switchMeshFormatToDrc(series.aortaMesh)
      }

      if (series.frames) {
        result.metadata = {
          width: series.frames.axial.width,
          height: series.frames.axial.height,
          depth: series.frames.axial.imageCount,
          pixelSpacing: series.frames.metadata.pixelSpacing
        }
      }

      res.json(result)

      return next()
    } catch (err) {
      return errorHandler.onError(req, res, next)(err)
    }
  },

  async updateArteryTree (req: request, res: response, next: Function) {
    const studyId = req.params.studyId
    const seriesId = req.params.seriesId
    const userId = req.user.sub
    const updateVessels = req.body.vessels

    try {
      await db.transaction(async trx => {
        const study = await studyService.getStudyByIdForUpdate(studyId, trx)
        if (!study) {
          throw new errors.NotFoundError(`Study ${studyId} is not found`)
        }
        let studyReview = await studyService.getStudyReviewByStudyIdForUpdate(studyId, trx) || {}
        if (studyReview.reviewedByMozart) {
          studyReview = await studyService.createStudyReviewFromMozartResult(studyId, userId, trx)
        }
        if (studyReview.reviewerId && studyReview.reviewerId !== userId) {
          throw new errors.ForbiddenError(`Can not edit. Study ${studyId} is currently under review by another reviewer.`)
        }

        const permission = getStudyPermission(study.status, studyReview.reviewerId, userId, req.user.roleName)
        if (!permission.canWrite) {
          return next(new errors.ForbiddenError(`Not allow to update this study`))
        }

        const series = studyReview.series[seriesId]
        if (!series) {
          throw new errors.NotFoundError(`Study ${studyId} series ${seriesId} is not found`)
        }

        let oldVesselDict = {}
        if (series.centralLineRefFile) {
          const oldArteryTree = await s3Helper.getS3Json(series.centralLineRefFile)
          oldVesselDict = _.keyBy(oldArteryTree.vessels, 'id')
        }

        if (!series.vessels) {
          series.vessels = {}
        }
        const newVesselDict = {}
        const vesselVertices = []
        for (const updateVessel of updateVessels) {
          const vessel = series.vessels[updateVessel.id]
          // database persisted vessel has no vertexes and wall
          const updatedPersistedVessel = _.omitBy({
            id: updateVessel.id,
            label: updateVessel.label,
            start: updateVessel.start,
            end: updateVessel.end,
            segments: updateVessel.segments,
            areas: updateVessel.areas,
            tools: updateVessel.tools
          }, _.isNil)

          if (!vessel && updateVessel.vertexes) {
            // If the vessel has vertexes but not in the series, mark it as not analyzed
            newVesselDict[updateVessel.id] = {
              ...updatedPersistedVessel,
              notAnalyzed: true // Mark as not analyzed
            }
            vesselVertices.push({
              id: updateVessel.id,
              vertexes: updateVessel.vertexes
            })
          } else {
            newVesselDict[updateVessel.id] = {
              ...vessel,
              ...updatedPersistedVessel
            }
            vesselVertices.push({
              id: updateVessel.id,
              vertexes: updateVessel.vertexes || (oldVesselDict[updateVessel.id] && oldVesselDict[updateVessel.id].vertexes)
            })
          }
        }

        // Remove the absent vessels
        series.vessels = newVesselDict

        const newArteryTreeS3Url = `s3://${dicomBucket.bucketName}/${studyId}/${shortId.generate()}/${seriesId}/artery_tree.json`
        series.centralLineRefFile = newArteryTreeS3Url

        await Promise.all([
          s3Helper.putS3Json(newArteryTreeS3Url, {
            vessels: vesselVertices
          }),
          seriesService.updateSeries(studyReview.id, seriesId, series, trx)
        ])

        const newVessels = _.map(vesselVertices, vv => {
          const v = newVesselDict[vv.id]
          v.vertexes = vv.vertexes
          return vesselUtil.getVesselSummary(v)
        })

        res.json({
          metadata: {
            width: series.frames.axial.width,
            height: series.frames.axial.height,
            depth: series.frames.axial.imageCount
          },
          vessels: newVessels,
          seriesVesselCombination: studyUtil.getSeriesVesselCombination(studyReview.series)
        })
        return next()
      })
    } catch (err) {
      return errorHandler.onError(req, res, next)(err)
    }
  },

  async manualHelper (req: request, res: response, next: Function) {
    const studyId = req.params.studyId
    const seriesId = req.params.seriesId
    const vertexes = req.body.vertexes

    try {
      const series = await seriesService.getSeriesById(studyId, seriesId)
      if (!series) {
        return next(new errors.NotFoundError(`Study ${studyId} series ${seriesId} is not found`))
      }

      let result = []
      const unknownPath = []

      let seriesFile = series.resampledSeriesExtractionRefFile
      let pixelSpacing = new Array(3).fill(series.frames.metadata.pixelSpacing)
      if (!seriesFile) {
        seriesFile = series.seriesExtractionRefFile
        pixelSpacing = series.metadata.voxelSpacing
      }
      joi.assert(pixelSpacing, joi.array().length(3).items(joi.number()))

      // send unknown path part to the manual helper
      // Keep the other parts in the cache
      for (let i = 0; i < vertexes.length - 1; i++) {
        const start = vertexes[i]
        const end = vertexes[i + 1]
        const part = manualHelperPathCache.get(manualHelperKey(start, end, seriesFile))
        if (_.isArray(part) && part.length > 0) {
          result = result.concat(part)
          if (i < vertexes.length - 2) {
            result.pop() // Remove the last end point
          }
        } else {
          unknownPath.push(start)
          if (i === vertexes.length - 2) {
            unknownPath.push(end)
          }
        }
      }

      const resp = await manualHelper(seriesFile, unknownPath, pixelSpacing)
        .catch(err => {
          // TODO: The default error unmarshall will break sentry log because there is a circular key reference
          if (err.response) {
            const newError = new Error(err.message)
            req.log.error(err.response.data)
            sentry.captureException(newError)
            sentry.captureMessage(err.response.data, 'error')
            throw newError
          } else {
            req.log.error(err)
            sentry.captureException(err)
            throw err
          }
        })

      const path = []
      let lastVertex = null
      for (const v of resp.vertexes) {
        if (!_.isEqual(v, lastVertex)) {
          path.push(v)
          lastVertex = v
        }
      }

      res.json({
        vertexes: result.concat(path)
      })
      return next()
    } catch (err) {
      return errorHandler.onError(req, res, next)(err)
    }
  },

  async getVesselById (req: request, res: response, next: Function) {
    const studyId = req.params.studyId
    const seriesId = req.params.seriesId
    const vesselId = req.params.vesselId
    const userId = req.user.sub

    try {
      const studyReview = await studyService.getMostRecentStudyReviewByStudyId(studyId)
      const { vessel, reviewerId, pixelSpacing } = await vesselService.getVesselById(studyReview.id, seriesId, vesselId)
      if (!vessel) {
        return next(new errors.NotFoundError(`Study ${studyId} series ${seriesId} vessel ${vesselId} is not found`))
      }

      const permission = getStudyPermission(studyReview.status, reviewerId, userId, req.user.roleName)
      if (!permission.canView) {
        return next(new errors.ForbiddenError(`Not allow to view the study ${studyId}`))
      }

      const [lumenWall, vesselWall, { statistics, plaqueAreas }] = await Promise.all([
        s3Helper.getS3Json(vessel.lumenWallRefFile).catch(err => {
          sentry.captureException(err)
          req.log.error(`Study ${studyId} series ${seriesId} vessel ${vesselId} - Fetch lumen wall failed`, err)
          return null
        }),
        s3Helper.getS3Json(vessel.vesselWallRefFile).catch(err => {
          sentry.captureException(err)
          req.log.error(`Study ${studyId} series ${seriesId} vessel ${vesselId} - Fetch vessel wall failed`, err)
          return null
        }),
        getStatisticsPlaqueAreas(studyId, seriesId, vesselId, vessel).catch(err => {
          // TODO: The default error unmarshall will break sentry log because there is a circular key reference
          if (err.response) {
            const newError = new Error(err.message)
            req.log.error(`Study ${studyId} series ${seriesId} vessel ${vesselId} - Fetch statistics failed`, newError)
            req.log.error(err.response.data)
            sentry.captureException(newError)
            sentry.captureMessage(err.response.data, 'error')
            throw newError
          } else {
            req.log.error(err)
            sentry.captureException(err)
            throw err
          }
        })
      ])

      vessel.lumenWall = lumenWall
      vessel.vesselWall = vesselWall
      vessel.statistics = statistics
      vessel.plaqueAreas = {}

      if (!_.isEmpty(plaqueAreas)) {
        for (const [key, value] of Object.entries(plaqueAreas)) {
          vessel.plaqueAreas[key] = {
            shape: vessel.shape,
            crossSectionData: value
          }
        }
      }

      res.json(vesselUtil.getVesselDetails(vessel, pixelSpacing))
      return next()
    } catch (err) {
      return errorHandler.onError(req, res, next)(err)
    }
  },

  async updateVessel (req: request, res: response, next: Function) {
    const studyId = req.params.studyId
    const seriesId = req.params.seriesId
    const vesselId = req.params.vesselId
    const userId = req.user.sub

    try {
      await db.transaction(async trx => {
        let studyReview = await studyService.getStudyReviewByStudyIdForUpdate(studyId, trx) || {}
        if (studyReview.reviewedByMozart) {
          studyReview = await studyService.createStudyReviewFromMozartResult(studyId, userId, trx)
        }

        let { vessel, reviewerId, pixelSpacing } = await vesselService.getVesselByIdForUpdate(studyReview.id, seriesId, vesselId, trx)
        if (!vessel) {
          return next(new errors.NotFoundError(`Study ${studyId} series ${seriesId} vessel ${vesselId} is not found`))
        }

        const permission = getStudyPermission(studyReview.status, reviewerId, userId, req.user.roleName)
        if (!permission.canWrite) {
          return next(new errors.ForbiddenError(`Not allow to update the study ${studyId}`))
        }

        let requireStatisticsUpdate = false
        vessel.reviewed = true
        if (vessel.pending) {
          return next(new errors.ForbiddenError(`Study ${studyId} series ${seriesId} vessel ${vesselId} is pending. Not allow to update the pending vessel.`))
        }

        if (req.body.label) {
          vessel.label = req.body.label
        }

        if (req.body.segments) {
          vessel.segments = _.map(req.body.segments, segment => {
            // If the segment has no id, generate one and save.
            return {
              ...segment,
              id: segment.id || `s_${shortId.generate()}`
            }
          })
        }

        if (req.body.start) {
          vessel.start = req.body.start
          requireStatisticsUpdate = true
        }

        if (req.body.end) {
          vessel.end = req.body.end
          requireStatisticsUpdate = true
        }

        let ctoChanged = false
        if (req.body.areas && req.body.areas.ALL) {
          const newCtoArea = vesselUtil.getCTOArea(req.body.areas.ALL)
          const oldCtoArea = vesselUtil.getCTOArea(_.get(vessel, 'areas.ALL'))
          ctoChanged = !_.isEqual(newCtoArea, oldCtoArea)

          vessel.areas = req.body.areas
          requireStatisticsUpdate = true
        }

        // `vessel` is for the response
        // `persistedVessel` is for the db persistence
        const persistedVessel = _.clone(vessel)

        const promises = []
        const wall = req.body.wall
        if (wall || ctoChanged) {
          vessel = {
            ...vessel,
            ...(await getWalls(vessel, wall))
          }
          delete vessel.variablesRefFile
          requireStatisticsUpdate = true
        }

        if (requireStatisticsUpdate) {
          // Wall update will not be sent back to the front end
          // So they are not set in the vessel
          promises.push(getStatisticsPlaqueAreas(studyId, seriesId, vesselId, vessel).catch(err => {
            // TODO: The default error unmarshall will break sentry log because there is a circular key reference
            if (err.response) {
              const newError = new Error(err.message)
              req.log.error(`Study ${studyId} series ${seriesId} vessel ${vesselId} - Fetch statistics failed`, newError)
              req.log.error(err.response.data)
              sentry.captureException(newError)
              sentry.captureMessage(err.response.data, 'error')
              throw newError
            } else {
              req.log.error(err)
              sentry.captureException(err)
              throw err
            }
          }))

          if (wall && !_.isEmpty(wall.lumen) && vessel.lumenWall) {
            const lumenWallS3Url = `s3://${dicomBucket.bucketName}/${studyId}/${shortId.generate()}/${seriesId}/${vessel.id}/lumen_wall.json`
            persistedVessel.lumenWallRefFile = lumenWallS3Url
            promises.push(s3Helper.putS3Json(lumenWallS3Url, vessel.lumenWall))
          }
          if (wall && !_.isEmpty(wall.vessel) && vessel.vesselWall) {
            const vesselWallS3Url = `s3://${dicomBucket.bucketName}/${studyId}/${shortId.generate()}/${seriesId}/${vessel.id}/vessel_wall.json`
            persistedVessel.vesselWallRefFile = vesselWallS3Url
            promises.push(s3Helper.putS3Json(vesselWallS3Url, vessel.vesselWall))
          }
        }

        if (!_.isEmpty(promises)) {
          const [{ statistics, plaqueAreas, variablesRefFile }] = await Promise.all(promises)
          vessel.statistics = statistics

          if (wall || ctoChanged) {
            vessel.plaqueAreas = {}
            for (const [key, value] of Object.entries(plaqueAreas)) {
              vessel.plaqueAreas[key] = {
                shape: vessel.shape,
                crossSectionData: value
              }
            }

            if (variablesRefFile) {
              persistedVessel.variablesRefFile = variablesRefFile
            }
          }
        }

        const results = await vesselService.updateVessel(studyReview.id, seriesId, vesselId, persistedVessel, trx)
        let updatedVessel = vesselUtil.getVesselDetails(vessel, pixelSpacing)
        if (!_.isEmpty(results)) {
          const dateUpdated = results[0]
          updatedVessel = {
            ...updatedVessel,
            dateUpdated
          }
        }
        res.json(updatedVessel)
        return next()
      })
    } catch (err) {
      return errorHandler.onError(req, res, next)(err)
    }
  },

  async preloadCentralPoint (req: request, res: response, next: Function) {
    const studyId = req.params.studyId
    const seriesId = req.params.seriesId
    const vesselId = req.params.vesselId

    try {
      const series = await seriesService.getSeriesById(studyId, seriesId)
      if (!series) {
        return next(new errors.NotFoundError(`Study ${studyId} series ${seriesId} vessel ${vesselId} is not found`))
      }

      const vessel = series.vessels[vesselId]
      if (!vessel) {
        return next(new errors.NotFoundError(`Study ${studyId} series ${seriesId} vessel ${vesselId} is not found`))
      }

      const payload = {
        centralLineSpacing: vessel.centralLineSpacing,
        seriesExtractionRefFile: series.seriesExtractionRefFile,
        transformMatrixRefFile: vessel.transformMatrix
      }
      centralPointCache.set(`${studyId}/${seriesId}/${vesselId}`, payload)

      await preloadCentralPoint(
        payload.seriesExtractionRefFile,
        payload.transformMatrixRefFile,
        payload.centralLineSpacing
      )

      res.send(204)
      return next()
    } catch (err) {
      return errorHandler.onError(req, res, next)(err)
    }
  },

  async adjustCentralPoint (req: request, res: response, next: Function) {
    const studyId = req.params.studyId
    const seriesId = req.params.seriesId
    const vesselId = req.params.vesselId
    const centralPoints = req.body.centralPoints

    try {
      let payload = centralPointCache.get(`${studyId}/${seriesId}/${vesselId}`)
      if (!payload) {
        const series = await seriesService.getSeriesById(studyId, seriesId)
        if (!series) {
          return next(new errors.NotFoundError(`Study ${studyId} series ${seriesId} vessel ${vesselId} is not found`))
        }

        const vessel = series.vessels[vesselId]
        if (!vessel) {
          return next(new errors.NotFoundError(`Study ${studyId} series ${seriesId} vessel ${vesselId} is not found`))
        }

        payload = {
          centralLineSpacing: vessel.centralLineSpacing || CENTRAL_LINE_SPACING,
          seriesExtractionRefFile: series.seriesExtractionRefFile,
          transformMatrixRefFile: vessel.transformMatrix
        }
        centralPointCache.set(`${studyId}/${seriesId}/${vesselId}`, payload)
      }

      const response = await adjustCentralPoint(
        centralPoints,
        payload.seriesExtractionRefFile,
        payload.transformMatrixRefFile,
        payload.centralLineSpacing)

      await streamToPromise(response.data.pipe(res))
      return next()
    } catch (err) {
      return errorHandler.onError(req, res, next)(err)
    }
  },

  async analyzeVesselById (req: request, res: response, next: Function) {
    const studyId = req.params.studyId
    const seriesId = req.params.seriesId
    const vesselId = req.params.vesselId
    const userId = req.user.sub
    const { vertexes } = req.body

    try {
      let task
      let vessel = {}
      let pixelSpacing = CENTRAL_LINE_SPACING
      await db.transaction(async trx => {
        const studyObject = await getStudyByIdForUpdate(studyId, trx)
        if (!studyObject) {
          throw new errors.NotFoundError(`Study ${studyId} is not found`)
        }

        let studyReview = await studyService.getStudyReviewByStudyIdForUpdate(studyId, trx) || {}
        if (studyReview.reviewedByMozart) {
          studyReview = await studyService.createStudyReviewFromMozartResult(studyId, userId, trx)
        }
        const permission = getStudyPermission(studyObject.status, studyReview.reviewerId, userId, req.user.roleName)
        if (!permission.canWrite) {
          return next(new errors.ForbiddenError(`Not allow to edit this study`))
        }

        // Attach new vertexes on the vessel
        if (!studyReview.series || !studyReview.series[seriesId]) {
          throw new errors.NotFoundError(`The study ${studyId} has no series ${seriesId}`)
        }

        if (_.isEmpty(studyReview.series[seriesId].vessels) ||
          !studyReview.series[seriesId].vessels[vesselId]) {
          throw new errors.NotFoundError(`The study ${studyId} series ${seriesId} has no vessel ${vesselId} not found`)
        }

        vessel = studyReview.series[seriesId].vessels[vesselId]
        vessel.notAnalyzed = true
        vessel.vertexes = vertexes
        pixelSpacing = studyReview.series[seriesId].frames.metadata.pixelSpacing

        task = await createTaskForVesselsCMPRAnalysis(studyId, studyReview, trx)
        delete vessel.vertexes

        // Add pending flag
        addPendingOnVessels(studyReview)
        await updateStudyReviewById(studyReview.id, {
          series: JSON.stringify(studyReview.series)
        }, trx)
      })

      // Send task message after the db transaction
      if (task) {
        await sendTaskSQSMessage(task)
      }

      res.json(vesselUtil.getVesselDetails(vessel, pixelSpacing))
      return next()
    } catch (err) {
      return errorHandler.onError(req, res, next)(err)
    }
  },

  async markVesselAsReviewed (req: request, res: response, next: Function) {
    const studyId = req.params.studyId
    const seriesId = req.params.seriesId
    const vesselId = req.params.vesselId
    const userId = req.user.sub

    try {
      await db.transaction(async trx => {
        const studyObject = await getStudyByIdForUpdate(studyId, trx)
        if (!studyObject) {
          throw new errors.NotFoundError(`Study ${studyId} is not found`)
        }

        let studyReview = await studyService.getStudyReviewByStudyIdForUpdate(studyId, trx) || {}
        if (studyReview.reviewedByMozart) {
          studyReview = await studyService.createStudyReviewFromMozartResult(studyId, userId, trx)
        }

        const permission = getStudyPermission(studyObject.status, studyReview.reviewerId, userId, req.user.roleName)
        if (!permission.canWrite) {
          return next(new errors.ForbiddenError(`Not allow to edit this study`))
        }

        // Attach new vertexes on the vessel
        if (!studyReview.series || !studyReview.series[seriesId]) {
          throw new errors.NotFoundError(`The study ${studyId} has no series ${seriesId}`)
        }

        if (_.isEmpty(studyReview.series[seriesId].vessels) ||
          !studyReview.series[seriesId].vessels[vesselId]) {
          throw new errors.NotFoundError(`The study ${studyId} series ${seriesId} has no vessel ${vesselId} not found`)
        }

        const vessel = studyReview.series[seriesId].vessels[vesselId]

        // Add reviewed flag
        vessel.reviewed = true
        await vesselService.updateVessel(studyReview.id, seriesId, vesselId, vessel, trx)
      })

      res.send(204)
      return next()
    } catch (err) {
      return errorHandler.onError(req, res, next)(err)
    }
  }
}
