// @flow
'use strict'

const joi = require('joi')

const server = require('../server')
const handler = require('./vessel-handler')
const { TOOLS } = require('./vessel-constant')
const { requireRole } = require('../lib/role')
const { ROLES, LAB_BASE_PATH } = require('../constant')

const STUDY_PATH = '/cases'
const VESSEL_PATH = '/vessels'
const SERIES_PATH = '/series'

const joiPoint3D = joi.object().keys({
  x: joi.number().min(0).required(),
  y: joi.number().min(0).required(),
  z: joi.number().min(0).required()
})

const joiVertexes = joi.array().min(2).items(joiPoint3D)

const joiPoint2D = joi.object().keys({
  x: joi.number().min(0).required(),
  y: joi.number().min(0).required()
})

const joiWall = joi.array().items(joi.object().keys({
  '0': joi.array().items(joiPoint2D),
  '22.5': joi.array().items(joiPoint2D),
  '45': joi.array().items(joiPoint2D),
  '67.5': joi.array().items(joiPoint2D),
  '90': joi.array().items(joiPoint2D),
  '112.5': joi.array().items(joiPoint2D),
  '135': joi.array().items(joiPoint2D),
  '157.5': joi.array().items(joiPoint2D)
}))

// TODO: add a better joi validator for each type of tools
const joiArea = joi.array().unique('type').items(joi.object().keys({
  path: joi.array().items(joiPoint2D),
  sections: joi.array().min(1).items(joi.object().keys({
    label: joi.string(),
    start: joi.number().integer().min(0),
    end: joi.number().integer().greater(joi.ref('start')),
    reason: joi.string().allow('')
  })),
  groups: joi.array().min(1).items(joi.object().keys({
    group: joi.array().length(5).items(joi.object().keys({
      position: joi.number().integer().min(0).required(),
      label: joi.string().valid('R1', 'P', 'O', 'D', 'R2').required()
    }))
  })),
  type: joi.string().valid(Object.values(TOOLS)).required()
}).xor('path', 'sections', 'groups').unknown(true))

const joiVessel = joi.object().keys({
  id: joi.string(),
  label: joi.string(),
  start: joi.number().integer().min(0),
  end: joi.number().integer().greater(joi.ref('start')),
  segments: joi.array().items(joi.object().keys({
    id: joi.string(),
    label: joi.string().required(),
    start: joi.number().integer().min(0).required(),
    end: joi.number().integer().greater(joi.ref('start')).required()
  })),
  areas: joi.object().keys({
    ALL: joiArea
  }),
  wall: joi.object().keys({
    startIndex: joi.number().integer().min(0),
    lumen: joiWall,
    vessel: joiWall
  }).or('lumen', 'vessel'),
  tools: joi.object(),
  vertexes: joiVertexes
}).or('id', 'label', 'start', 'end', 'segments', 'areas', 'wall', 'tools', 'vertexes')

/**
 * Get series vessels artery tree
 */
server.get({
  path: `${LAB_BASE_PATH}${STUDY_PATH}/:studyId${SERIES_PATH}/:seriesId${VESSEL_PATH}`,
  validation: {
    schema: joi.object().keys({
      params: joi.object().keys({
        studyId: joi.string().guid().required(),
        seriesId: joi.string().min(1).required()
      }).required()
    })
  }
}, [
  requireRole([
    ROLES.SYSTEM_ADMIN,
    ROLES.LAB_ADMIN,
    ROLES.LAB_USER
  ]),
  handler.getArteryTree
])

/**
 * Update artery tree
 */
server.put({
  path: `${LAB_BASE_PATH}${STUDY_PATH}/:studyId${SERIES_PATH}/:seriesId${VESSEL_PATH}`,
  validation: {
    schema: joi.object().keys({
      params: joi.object().keys({
        studyId: joi.string().guid().required(),
        seriesId: joi.string().min(1).required()
      }).required(),
      body: joi.object().keys({
        vessels: joi.array().min(1).items(joiVessel)
      }).required()
    })
  }
}, [
  requireRole([
    ROLES.SYSTEM_ADMIN,
    ROLES.LAB_ADMIN,
    ROLES.LAB_USER
  ]),
  handler.updateArteryTree
])

/**
 * Manual helper for semi-auto/manual vessel creation
 */
server.post({
  path: `${LAB_BASE_PATH}${STUDY_PATH}/:studyId${SERIES_PATH}/:seriesId${VESSEL_PATH}/manualhelper`,
  validation: {
    schema: joi.object().keys({
      params: joi.object().keys({
        studyId: joi.string().guid().required(),
        seriesId: joi.string().min(1).required()
      }).required(),
      body: joi.object().keys({
        vertexes: joiVertexes.required()
      })
    })
  }
}, [
  requireRole([
    ROLES.SYSTEM_ADMIN,
    ROLES.LAB_ADMIN,
    ROLES.LAB_USER
  ]),
  handler.manualHelper
])

/**
 * Update a vessel
 */
server.put({
  path: `${LAB_BASE_PATH}${STUDY_PATH}/:studyId${SERIES_PATH}/:seriesId${VESSEL_PATH}/:vesselId`,
  validation: {
    schema: joi.object().keys({
      params: joi.object().keys({
        studyId: joi.string().guid().required(),
        seriesId: joi.string().min(1).required(),
        vesselId: joi.string().required()
      }).required(),
      body: joiVessel.required()
    })
  }
}, [
  requireRole([
    ROLES.SYSTEM_ADMIN,
    ROLES.LAB_ADMIN,
    ROLES.LAB_USER
  ]),
  handler.updateVessel
])

/**
 * Get vessel information
 */
server.get({
  path: `${LAB_BASE_PATH}${STUDY_PATH}/:studyId${SERIES_PATH}/:seriesId${VESSEL_PATH}/:vesselId`,
  validation: {
    schema: joi.object().keys({
      params: joi.object().keys({
        studyId: joi.string().guid().required(),
        seriesId: joi.string().required(),
        vesselId: joi.string().required()
      }).required()
    })
  }
}, [
  requireRole([
    ROLES.SYSTEM_ADMIN,
    ROLES.LAB_ADMIN,
    ROLES.LAB_USER
  ]),
  handler.getVesselById
])

/**
 * Central point adjustment preload
 */
server.post({
  path: `${LAB_BASE_PATH}${STUDY_PATH}/:studyId${SERIES_PATH}/:seriesId${VESSEL_PATH}/:vesselId/central-points-preload`,
  validation: {
    schema: joi.object().keys({
      params: joi.object().keys({
        studyId: joi.string().guid().required(),
        seriesId: joi.string().required(),
        vesselId: joi.string().required()
      }).required()
    })
  }
}, [
  requireRole([
    ROLES.SYSTEM_ADMIN,
    ROLES.LAB_ADMIN,
    ROLES.LAB_USER
  ]),
  handler.preloadCentralPoint
])

/**
 * Central point adjustment
 */
server.put({
  path: `${LAB_BASE_PATH}${STUDY_PATH}/:studyId${SERIES_PATH}/:seriesId${VESSEL_PATH}/:vesselId/central-points`,
  validation: {
    schema: joi.object().keys({
      params: joi.object().keys({
        studyId: joi.string().guid().required(),
        seriesId: joi.string().required(),
        vesselId: joi.string().required()
      }).required(),
      body: joi.object().keys({
        centralPoints: joi.array().min(1).items(joi.object().keys({
          x: joi.number(),
          y: joi.number(),
          transformMatrix: joi.array().length(3).items(
            joi.array().length(3).items(joi.number())
          ),
          zIndex: joi.number()
        })).required()
      })
    })
  }
}, [
  requireRole([
    ROLES.SYSTEM_ADMIN,
    ROLES.LAB_ADMIN,
    ROLES.LAB_USER
  ]),
  handler.adjustCentralPoint
])

/**
 * Full analyze vessel after central point confirm
 */
server.put({
  path: `${LAB_BASE_PATH}${STUDY_PATH}/:studyId${SERIES_PATH}/:seriesId${VESSEL_PATH}/:vesselId/analysis`,
  validation: {
    schema: joi.object().keys({
      params: joi.object().keys({
        studyId: joi.string().guid().required(),
        seriesId: joi.string().required(),
        vesselId: joi.string().required()
      }).required(),
      body: joi.object().keys({
        vertexes: joiVertexes.required()
      }).required()
    })
  }
}, [
  requireRole([
    ROLES.SYSTEM_ADMIN,
    ROLES.LAB_ADMIN,
    ROLES.LAB_USER
  ]),
  handler.analyzeVesselById
])

/**
 * Mark the vessel as reviewed
 */
server.put({
  path: `${LAB_BASE_PATH}${STUDY_PATH}/:studyId${SERIES_PATH}/:seriesId${VESSEL_PATH}/:vesselId/reviewed`,
  validation: {
    schema: joi.object().keys({
      params: joi.object().keys({
        studyId: joi.string().guid().required(),
        seriesId: joi.string().required(),
        vesselId: joi.string().required()
      }).required()
    }).required()
  }
}, [
  requireRole([
    ROLES.SYSTEM_ADMIN,
    ROLES.LAB_ADMIN,
    ROLES.LAB_USER
  ]),
  handler.markVesselAsReviewed
])
